; FPU_RAMP.ASM
; ------------
; uFMOD public source code release. (MiniFMOD v1.70 based.) Provided as-is.
; Original Visual C++ source optimized and debugged by Asterix and Quantum.
; uFMOD sourcecode is (c)opyleft [WASM.RU]. All rights reserved & reversed.
; MiniFMOD sourcecode is (c)opyright Firelight Technologies (R), 2000-2005.
; uFMOD sourcecode is in no way representative of FMOD 3 source.
; This source must not be redistributed without this notice.

.486
.model flat,stdcall
option casemap:none

include sound.inc

mix_volumerampsteps EQU FSOUND_MixRate * 128 / 44100 ; at 44.1KHz

.DATA
; made global to free ebp
mix_1over255       dd 3.9215689e-3 ; 1.0f / 255.0f
mix_1over2gig      dd 4.6566129e-10 ; 1.0f / 2147483648.0f
inv_rampsteps      dd 3.0637257e-5 ; 1.0f / mix_volumerampsteps / 255.0f

.DATA?
; made global to free ebp
mix_mixbuffend     dd ?
mix_mixbuffptr     dd ?
mix_endflag        dd ?
mix_sptr           dd ?
mix_cptr           dd ?
mix_count          dd ?
mix_samplebuff     dd ?
mix_count_old      dd ?
mix_rampleftvol    dd ?
mix_ramprightvol   dd ?
mix_rampcount      dd ?
mix_rampspeedleft  dd ?
mix_rampspeedright dd ?
mix_leftvol        dd ?
mix_rightvol       dd ?

.CODE

; DESCRIPTION: Mixer ramping (for internal usage)
; PARAMETERS: pointer to mixer (mixptr) and it's length (len)
uFMOD_FPU_Ramp PROC mixptr:DWORD,len:DWORD
option prologue:none
option epilogue:none
LOCAL count:DWORD
   push ebp
   mov ebp,esp
   push eax ; make some space for a local variable
   mov eax,len
   mov ecx,mixptr
   lea eax,[ecx + eax * 8]
   mov mix_mixbuffend,eax ; mix_mixbuffend = mixptr + (len << 3)
   push ebx
   push esi
   push edi
   ; LOOP THROUGH CHANNELS
   mov count,OFFSET FSOUND_Channel
loop_ch:
   mov eax,mixptr
   mov mix_mixbuffptr,eax
   mov ecx,count
   mov mix_cptr,ecx
   mov ebx,[ecx + FSOUND_CHANNEL.sptr] ; load the correct SAMPLE pointer for this channel
   mov mix_sptr,ebx ; store sample pointer
   mov edi,len
   push ebp
   test ebx,ebx ; if(!sptr) then skip this channel!
   jz MixExit
   mov edx,[ebx + FSOUND_SAMPLE.buff] ; get pointer to sample buffer
   shr edx,1
   mov mix_samplebuff,edx
   ; the following code sets up a mix counter. it sees what will happen first, will the output buffer
   ; end be reached first? or will the end of the sample be reached first? whatever is smallest will
   ; be the mixcount.
   ; first base mixcount on size of OUTPUT BUFFER (in samples not bytes)
CalculateLoopCount:
   mov mix_count,edi
   mov esi,[ecx + FSOUND_CHANNEL.mixpos]
   xor ebp,ebp
   mov mix_endflag,ebp ; set a flag to say mixing will end when end of output buffer is reached
   cmp [ecx + FSOUND_CHANNEL.speeddir],ebp ; compare to FSOUND_MIXDIR_FORWARDS
   mov eax,[ecx + FSOUND_CHANNEL.mixposlo]
   jne samplesleftbackwards
   ; work out how many samples left from mixpos to loop end
   mov edx,[ebx + FSOUND_SAMPLE.loopstart]
   add edx,[ebx + FSOUND_SAMPLE.looplen]
   cmp esi,edx
   jle submixpos
   mov edx,[ebx+FSOUND_SAMPLE._length]
submixpos:
   sub edx,esi ; edx = samples left (loopstart+looplen-mixpos)
   neg eax
   sbb edx,ebp ; ebp == 0
   jmp samplesleftfinish
samplesleftbackwards:
   ; work out how many samples left from mixpos to loop start
   mov edx,[ecx + FSOUND_CHANNEL.mixpos]
   sub edx,[ebx + FSOUND_SAMPLE.loopstart]
samplesleftfinish:
   ; edx:eax now contains number of samples left to mix
   cmp edx,1000000h
   jae staywithoutputbuffend
   shrd eax,edx,8
   shr edx,8
   ; now samples left = EDX:EAX -> hhhhhlll
   mov ebp,[ecx + FSOUND_CHANNEL.speedhi]
   mov edi,[ecx + FSOUND_CHANNEL.speedlo]
   shr edi,8
   shl ebp,24
   or ebp,edi
   jnz speedok ; do a paranoid divide by 0 check
   mov ebp,017C6Fh ; 100Hz
speedok:
   div ebp
   test edx,edx ; if fractional 16-bit part is zero we must add an extra carry number
   jz dontaddbyte ; to the resultant in EDX:EAX
   inc eax
dontaddbyte: ; we must remove the fractional part of the multiply by shifting EDX:EAX
   cmp eax,mix_count
   ja staywithoutputbuffend
   mov mix_count,eax
   ; set a flag to say mix will end when end of output buffer is reached
   inc mix_endflag
staywithoutputbuffend:
   ; VOLUME RAMP SETUP
   ; Reasons to ramp
   ; 1 volume change
   ; 2 sample starts (just treat as volume change - 0 to volume)
   ; 3 sample ends (ramp last n number of samples from volume to 0)
   ; now if the volume has changed, make end condition equal a volume ramp
   xor eax,eax
   mov mix_rampspeedleft,eax ; clear out volume ramp
   mov mix_rampspeedright,eax ; clear out volume ramp
   mov mix_rampcount,eax
   cmp [ecx + FSOUND_CHANNEL.ramp_count],eax
   mov eax,mix_count
   mov mix_count_old,eax ; remember mix count before modifying it
   mov eax,[ecx + FSOUND_CHANNEL.leftvolume]
   je volumerampstart
   ; if it tries to continue an old ramp, but the target has changed,
   ; set up a new ramp
   cmp eax,[ecx + FSOUND_CHANNEL.ramp_lefttarget]
   jne volumerampstart
   mov edx,[ecx + FSOUND_CHANNEL.ramp_righttarget]
   cmp [ecx + FSOUND_CHANNEL.rightvolume],edx
   jne volumerampstart
   ; restore old ramp
   mov eax,[ecx + FSOUND_CHANNEL.ramp_count]
   mov mix_rampcount,eax
   mov eax,[ecx + FSOUND_CHANNEL.ramp_leftspeed]
   mov mix_rampspeedleft,eax
   mov eax,[ecx + FSOUND_CHANNEL.ramp_rightspeed]
   mov mix_rampspeedright,eax
   jmp novolumerampR
volumerampstart:
   mov edx,[ecx + FSOUND_CHANNEL.ramp_leftvolume]
   mov [ecx + FSOUND_CHANNEL.ramp_lefttarget],eax
   sub eax,edx
   jz novolumerampL
   mov mix_rampspeedleft,eax
   fild mix_rampspeedleft
   fmul inv_rampsteps
   fstp mix_rampspeedleft
   mov eax,mix_rampspeedleft
   mov [ecx + FSOUND_CHANNEL.ramp_leftspeed],eax
   mov eax,mix_volumerampsteps
   mov mix_rampcount,eax
novolumerampL:
   mov eax,[ecx + FSOUND_CHANNEL.rightvolume]
   mov edx,[ecx + FSOUND_CHANNEL.ramp_rightvolume]
   mov [ecx + FSOUND_CHANNEL.ramp_righttarget],eax
   sub eax,edx
   jz novolumerampR
   mov mix_rampspeedright,eax
   fild mix_rampspeedright
   fmul inv_rampsteps
   fstp mix_rampspeedright
   mov eax,mix_rampspeedright
   mov [ecx + FSOUND_CHANNEL.ramp_rightspeed],eax
   mov eax,mix_volumerampsteps
   mov mix_rampcount,eax
novolumerampR:
   mov eax,mix_rampcount
   test eax,eax
   jle volumerampend
   mov [ecx + FSOUND_CHANNEL.ramp_count],eax
   cmp mix_count,eax
   jbe volumerampend ; dont clamp mixcount
   mov mix_count,eax
volumerampend:
   ; SET UP VOLUME MULTIPLIERS
   ; right volume
   mov eax,[ecx + FSOUND_CHANNEL.rightvolume]
   mov mix_rampleftvol,eax
   fild mix_rampleftvol
   fmul mix_1over255
   fstp mix_rightvol
   ; left volume
   mov eax,[ecx + FSOUND_CHANNEL.leftvolume]
   mov mix_rampleftvol,eax
   fild mix_rampleftvol
   fmul mix_1over255
   fstp mix_leftvol
   ; right ramp volume
   mov eax,[ecx + FSOUND_CHANNEL.ramp_rightvolume]
   mov mix_rampleftvol,eax
   fild mix_rampleftvol
   fmul mix_1over255 ; convert from 24:8 to 0-255 and make 0-1.0f
   fstp mix_ramprightvol
   ; left ramp volume
   mov eax,[ecx + FSOUND_CHANNEL.ramp_leftvolume]
   mov mix_rampleftvol,eax
   fild mix_rampleftvol
   fmul mix_1over255 ; convert from 24:8 to 0-255 and make 0-1.0f
   fstp mix_rampleftvol
   ; SET UP ALL OF THE REGISTERS HERE FOR THE INNER LOOP
   ; ebx = speed low
   ; edx = speed high
   ; esi = mixpos
   ; edi = destination pointer
   ; ebp = mixpos low
   mov ebx,[ecx + FSOUND_CHANNEL.speedlo]
   mov ebp,[ecx + FSOUND_CHANNEL.mixposlo]
   mov esi,[ecx + FSOUND_CHANNEL.mixpos]
   mov edi,mix_mixbuffptr ; point to 16bit output stream
   cmp [ecx + FSOUND_CHANNEL.speeddir],FSOUND_MIXDIR_FORWARDS
   mov edx,[ecx + FSOUND_CHANNEL.speedhi]
   je NoChangeSpeed
   not ebx
   not edx
   inc ebx
   adc edx,0
NoChangeSpeed:
   ; 16 BIT NORMAL FUNCTIONS
   mov eax,mix_samplebuff
   add esi,eax
   ; ecx = loop counter
   mov ecx,mix_count
   cmp mix_rampcount,0
   jne MixLoopStart16_2
   shr ecx,1
   jz MixLoopStart16 ; no groups of 2 samples to mix!
   ; START
   call deltas ; compute volume deltas
   fxch st(1) ; [0]interp1 [1]interp2 [2]samp1 [3]samp2
   faddp st(2),st(0) ; [0]interp2 [1]newsamp1 [2]samp2
   fld st(1) ; [0]newsamp1 [1]interp2 [2]newsamp1 [3]samp2
   fmul mix_leftvol ; [0]newsampL1 [1]interp2 [2]newsamp1 [3]samp2
   fxch st(1) ; [0]interp2 [1]newsampL1 [2]newsamp1 [3]samp2
   faddp st(3),st(0) ; [0]newsampL1 [1]newsamp1 [2]newsamp2
   jmp MixLoopUnroll16CoilEntry
MixLoopUnroll16:
   call deltas
   fxch st(3)
   fxch st(2) ; [0]samp1 [1]interp1 [2]samp2 [3]interp2 [4]finalR2 [5]finalR1 [6]finalL2
   faddp st(1),st(0) ; [0]newsamp1 [1]samp2 [2]interp2 [3]finalR2 [4]finalR1 [5]finalL2
   fxch st(4) ; [0]finalR1 [1]samp2 [2]interp2 [3]finalR2 [4]newsamp1 [5]finalL2
   fstp dword ptr [edi-28] ; [0]samp2 [1]interp2 [2]finalR2 [3]newsamp1 [4]finalL2
   fxch st(4) ; [0]finalL2 [1]interp2 [2]finalR2 [3]newsamp1 [4]samp2
   fstp dword ptr [edi-24] ; [0]interp2 [1]finalR2 [2]newsamp1 [3]samp2
   fld st(2) ; [0]newsamp1 [1]interp2 [2]finalR2 [3]newsamp1 [4]samp2
   fmul mix_leftvol ; [0]newsampL1 [1]interp2 [2]finalR2 [3]newsamp1 [4]samp2
   fxch st(1) ; [0]interp2 [1]newsampL1 [2]finalR2 [3]newsamp1 [4]samp2
   faddp st(4),st(0) ; [0]newsampL1 [1]finalR2 [2]newsamp1 [3]newsamp2
   fxch st(1) ; [0]finalR2 [1]newsampL1 [2]newsamp1 [3]newsamp2
   fstp dword ptr [edi-20] ; [0]newsampL1 [1]newsamp1 [2]newsamp2
MixLoopUnroll16CoilEntry:
   shl ebp,1
   add ebp,ebx
   adc esi,edx
   fxch st(1) ; [0]newsamp1 [1]newsampL1 [2]newsamp2
   fmul mix_rightvol ; [0]newsampR1 [1]newsampL1 [2]newsamp2
   fxch st(2) ; [0]newsamp2 [1]newsampL1 [2]newsampR1
   fld st(0) ; [0]newsamp2 [1]newsamp2 [2]newsampL1 [3]newsampR1
   fmul mix_leftvol ; [0]newsampL2 [1]newsamp2 [2]newsampL1 [3]newsampR1
   fxch st(1) ; [0]newsamp2 [1]newsampL2 [2]newsampL1 [3]newsampR1
   fmul mix_rightvol ; [0]newsampR2 [1]newsampL2 [2]newsampL1 [3]newsampR1
   fxch st(2) ; [0]newsampL1 [1]newsampL2 [2]newsampR2 [3]newsampR1
   fadd dword ptr [edi-16] ; [0]finalL1 [1]newsampL2 [2]newsampR2 [3]newsampR1
   fxch st(3) ; [0]newsampR1 [1]newsampL2 [2]newsampR2 [3]finalL1
   fadd dword ptr [edi-12] ; [0]finalR1 [1]newsampL2 [2]newsampR2 [3]finalL1
   fxch st(1) ; [0]newsampL2 [1]finalR1 [2]newsampR2 [3]finalL1
   fadd dword ptr [edi-8] ; [0]finalL2 [1]finalR1 [2]newsampR2 [3]finalL1
   fxch st(3) ; [0]finalL1 [1]finalR1 [2]newsampR2 [3]finalL2
   fstp dword ptr [edi-16] ; [0]finalR1 [1]newsampR2 [2]finalL2
   fxch st(1) ; [0]newsampR2 [1]finalR1 [2]finalL2
   fadd dword ptr [edi-4] ; [0]finalR2 [1]finalR1 [2]finalL2
   loop MixLoopUnroll16
   fxch st(1) ; [0]finalR1 [1]finalR2 [2]finalL2
   fstp dword ptr [edi-12] ; [0]finalR2 [1]finalL2
   fxch st(1) ; [0]finalL2 [1]finalR2
   fstp dword ptr [edi-8] ; [0]finalR2
   fstp dword ptr [edi-4]
   ; MIX 16BIT, ROLLED
MixLoopStart16:
   mov ecx,mix_count
   and ecx,1
MixLoopStart16_2:
   test ecx,ecx ; if(!count) dont enter the mix loop
   jz MixLoopEnd16
   fld mix_rampspeedleft ; [0]rampspeedL
   fld mix_rampleftvol
   fld mix_ramprightvol
   fld mix_rampspeedright
MixLoop16:
   shr ebp,1 ; make 31bit coz fpu only loads signed values
   mov mix_rampleftvol,ebp
   fild word ptr [esi+esi+2] ; [0]samp1+1 [1]rvol [2]lvol [3]rampspeedR [4]rampspeedL
   add edi,8
   fild word ptr [esi+esi] ; [0]samp1 [2]samp1+1 [3]rvol [4]lvol [5]rampspeedR [6]rampspeedL
   shl ebp,1
   fsub st(1),st(0)
   fild mix_rampleftvol ; [0]ifrac [1]samp1 [2]samp1+1 [3]rvol [4]lvol [5]rampspeedR [6]rampspeedL
   add ebp,ebx ; add speed low to mixpos low
   adc esi,edx ; add upper portion of speed plus carry
   fmul mix_1over2gig ; [0]ifrac [1]samp1 [2]samp1+1 [3]rvol [4]lvol [5]rampspeedR [6]rampspeedL
   fmulp st(2),st(0) ; [0]sample [1]interp [2]rvol [3]lvol [4]rampspeedR [5]rampspeedL
   faddp st(1),st(0) ; [0]newsamp [1]rvol [2]lvol [3]rampspeedR [4]rampspeedL
   fld st(0) ; [0]newsamp [1]newsamp [2]rvol [3]lvol [4]rampspeedR [5]rampspeedL
   fmul st(0),st(4)
   fadd dword ptr [edi-8]
   fstp dword ptr [edi-8]
   fmul st(0),st(2)
   fadd dword ptr [edi-4] ; [0]finalR [1]lvol [2]finalL [3]rvol [4]rampspeedR [5]rampspeedL
   fstp dword ptr [edi-4]
   fadd st(1),st(0)
   fxch st(2)
   fadd st(0),st(3)
   fxch st(2)
   loop MixLoop16
   fstp mix_rampspeedright
   fdiv mix_1over255
   fxch st(2)
   fstp mix_rampspeedleft
   fdiv mix_1over255
   fistp mix_rampleftvol
   fistp mix_ramprightvol
MixLoopEnd16:
   sub esi,mix_samplebuff
   mov mix_mixbuffptr,edi ; update the new mixbuffer pointer
   mov eax,mix_mixbuffend ; find out how many OUTPUT samples left to mix
   sub eax,edi
   shr eax,3
   xchg eax,edi ; edi now holds # of samples left
   ; DID A VOLUME RAMP JUST HAPPEN
   cmp mix_rampcount,0
   mov ecx,mix_cptr ; load ecx with channel pointer
   mov [ecx + FSOUND_CHANNEL.mixpos],esi
   mov [ecx + FSOUND_CHANNEL.mixposlo],ebp
   mov ebx,mix_sptr ; load ebx with sample pointer
   je DoOutputbuffEnd ; no, no ramp
   mov eax,mix_rampleftvol
   mov [ecx + FSOUND_CHANNEL.ramp_leftvolume],eax
   mov eax,mix_ramprightvol
   mov [ecx + FSOUND_CHANNEL.ramp_rightvolume],eax
   mov edx,mix_rampcount
   mov eax,mix_count
   sub edx,eax
   mov mix_rampcount,edx
   mov [ecx + FSOUND_CHANNEL.ramp_count],edx
   cdq
   mov mix_rampspeedleft,edx ; clear out volume ramp
   mov mix_rampspeedright,edx ; clear out volume ramp
   ; if(!rampcount) a ramp has FINISHED, so finish the rest of the mix
   jnz DoOutputbuffEnd
   ; clear out the ramp speeds
   mov [ecx + FSOUND_CHANNEL.ramp_leftspeed],edx
   mov [ecx + FSOUND_CHANNEL.ramp_rightspeed],edx
   ; clamp the 2 volumes together in case the speed wasnt accurate enough!
   mov edx,[ecx + FSOUND_CHANNEL.leftvolume]
   mov [ecx + FSOUND_CHANNEL.ramp_leftvolume],edx
   mov edx,[ecx + FSOUND_CHANNEL.rightvolume]
   mov [ecx + FSOUND_CHANNEL.ramp_rightvolume],edx
   ; is it 0 because ramp ended only? or both ended together?
   ; if sample ended together with ramp... problems... loop isn't handled
   cmp mix_count_old,eax ; ramp and output mode ended together
   je DoOutputbuffEnd
   ; start again and continue rest of mix
   test edi,edi
   jnz CalculateLoopCount ; dont start again if nothing left
DoOutputbuffEnd:
   cmp mix_endflag,0
   je MixExit
   ; SWITCH ON LOOP MODE TYPE
   mov al,[ebx + FSOUND_SAMPLE.loopmode]
   test al,FSOUND_LOOP_NORMAL ; check for normal loop
   jz CheckBidiLoop
   mov eax,[ebx + FSOUND_SAMPLE.loopstart]
   add eax,[ebx + FSOUND_SAMPLE.looplen]
rewindsample:
   sub esi,[ebx + FSOUND_SAMPLE.looplen]
   cmp esi,eax
   jae rewindsample
   mov [ecx + FSOUND_CHANNEL.mixpos],esi
   test edi,edi
   jz MixExit
   jmp CalculateLoopCount
CheckBidiLoop:
   test al,FSOUND_LOOP_BIDI
   jz NoLoop
   cmp [ecx + FSOUND_CHANNEL.speeddir],FSOUND_MIXDIR_FORWARDS
   je BidiForward
BidiBackwards:
   mov eax,[ebx + FSOUND_SAMPLE.loopstart]
   dec eax
   cdq
   dec edx
   sub edx,ebp
   sbb eax,esi
   xchg eax,esi
   mov ebp,edx ; esi:ebp = loopstart - mixpos
   mov eax,[ebx + FSOUND_SAMPLE.loopstart]
   add esi,eax ; esi:ebp += loopstart
   and [ecx + FSOUND_CHANNEL.speeddir],0 ; set FSOUND_MIXDIR_FORWARDS
   add eax,[ebx + FSOUND_SAMPLE.looplen]
   cmp esi,eax
   jl BidiFinish
BidiForward:
   mov eax,[ebx + FSOUND_SAMPLE.loopstart]
   add eax,[ebx + FSOUND_SAMPLE.looplen]
   push eax
   neg ebp
   sbb eax,esi
   xchg eax,esi ; esi:ebp = loopstart+looplen - mixpos
   pop eax
   dec eax
   cdq
   dec edx
   add ebp,edx
   adc esi,eax
   inc [ecx + FSOUND_CHANNEL.speeddir] ; set FSOUND_MIXDIR_BACKWARDS
   cmp esi,[ebx + FSOUND_SAMPLE.loopstart]
   jl BidiBackwards
BidiFinish:
   mov [ecx + FSOUND_CHANNEL.mixpos],esi
   mov [ecx + FSOUND_CHANNEL.mixposlo],ebp
   test edi,edi
   jz MixExit
   jmp CalculateLoopCount
NoLoop:
   xor eax,eax
   mov [ecx + FSOUND_CHANNEL.sptr],eax ; clear the sample pointer out
   mov [ecx + FSOUND_CHANNEL.mixposlo],eax
   mov [ecx + FSOUND_CHANNEL.mixpos],eax ; reset mixpos for next time around
MixExit:
   pop ebp ; restore ebp
   add count,SIZEOF FSOUND_CHANNEL
   cmp count,OFFSET FSOUND_Channel + SIZEOF FSOUND_CHANNEL * 64
   jl loop_ch ; iterate 64 times
   pop edi
   pop esi
   pop ebx
   leave
   retn
deltas:
   ; volume deltas
   shr ebp,1 ; make 31bit coz fpu only loads signed values
   mov eax,esi
   mov mix_mixbuffptr,ebp
   shl ebp,1
   add ebp,ebx
   adc esi,edx
   fild word ptr [esi+esi] ; samp2
   add edi,16
   shr ebp,1
   fild word ptr [eax+eax] ; samp1
   fild dword ptr mix_mixbuffptr ; ifrac1
   mov mix_mixbuffptr,ebp
   fmul mix_1over2gig ; ffrac1
   fild word ptr [eax+eax+2] ; samp1+1
   fsub st(0),st(2)
   fmulp st(1),st(0) ; ifrac2
   fild dword ptr mix_mixbuffptr
   fmul mix_1over2gig ; ffrac2
   fild word ptr [esi+esi+2] ; samp2+1
   fsub st(0),st(4) ; delta2
   fmulp st(1),st(0) ; interp2
   retn
uFMOD_FPU_Ramp ENDP

end
