// MUSIC.H
// -------
// uFMOD public source code release. (MiniFMOD v1.70 based.) Provided as-is.
// Original Visual C++ source optimized and debugged by Asterix and Quantum.
// uFMOD sourcecode is (c)opyleft [WASM.RU]. All rights reserved & reversed.
// MiniFMOD sourcecode is (c)opyright Firelight Technologies (R), 2000-2005.
// uFMOD sourcecode is in no way representative of FMOD 3 source.
// This source must not be redistributed without this notice.

#ifndef _MUSIC_H
#define _MUSIC_H

#include "sound.h"

#define FMUSIC_KEYOFF 255
#define FMUSIC_FREQ 1
#define FMUSIC_VOLUME 2
#define FMUSIC_PAN 4
#define FMUSIC_TRIGGER 8
#define FMUSIC_STOP 32
#define FMUSIC_ENVELOPE_OFF 0
#define FMUSIC_ENVELOPE_ON 1
#define FMUSIC_ENVELOPE_SUSTAIN 2
#define FMUSIC_ENVELOPE_LOOP 4

// Single note type - contains info on 1 note in a pattern
typedef struct{
   unsigned char note; // note to play at (0-133) 132=none,133=keyoff
   unsigned char number; // sample being played (0-99)
   unsigned char volume; // volume column value (0-64)  255=no volume
   unsigned char effect; // effect number (0-1Ah)
   unsigned char eparam; // effect parameter (0-255)
}FMUSIC_NOTE;

// pattern data type 
typedef struct{
   int rows;
   FMUSIC_NOTE* data;
}FMUSIC_PATTERN;

// Multi sample extended instrument
typedef struct FMUSIC_INSTRUMENT{
   int numsamples; // number of samples in this instrument
   FSOUND_SAMPLE* sample[16]; // 16 samples per instrument
   unsigned char keymap[96]; // sample keymap assignments
   unsigned short VOLPoints[24]; // volume envelope points
   unsigned short PANPoints[24]; // panning envelope points

   // 16 bytes read here in the loader for size, dont alter order or insert!!!
   unsigned char VOLnumpoints; // number of volume envelope points
   unsigned char PANnumpoints; // number of panning envelope points
   unsigned char VOLsustain; // volume sustain point
   unsigned char VOLLoopStart; // volume envelope loop start
   unsigned char VOLLoopEnd; // volume envelope loop end
   unsigned char PANsustain; // panning sustain point
   unsigned char PANLoopStart; // panning envelope loop start
   unsigned char PANLoopEnd; // panning envelope loop end
   unsigned char VOLtype; // type of envelope,bit 0:On 1:Sustain 2:Loop
   unsigned char PANtype; // type of envelope,bit 0:On 1:Sustain 2:Loop
   unsigned char VIBtype; // instrument Vibrato type
   unsigned char VIBsweep; // time it takes for vibrato to fully kick in
   unsigned char VIBdepth; // depth of vibrato
   unsigned char VIBrate; // rate of vibrato
   unsigned short VOLfade; // fade out value
}FMUSIC_INSTRUMENT;

// Channel type - contains information on a mod channel
typedef struct FMUSIC_CHANNEL{
   unsigned char note; // last note set in channel
   unsigned char samp; // last sample set in channel
   unsigned char notectrl;
   FSOUND_CHANNEL* cptr; // pointer to FSOUND system mixing channel
   FSOUND_SAMPLE* sptr; // pointer to FSOUND system sample
   int freq; // current mod frequency period for this channel
   int volume; // current mod volume for this channel
   int pan; // current mod pan for this channel
   int voldelta; // delta for volume commands.. tremolo/tremor etc
   int freqdelta; // delta for frequency commands.. vibrato/arpeggio etc
   int envvoltick; // tick counter for envelope position
   int envvolpos; // envelope position
   int envvolfrac; // fractional interpolated envelope volume
   int envvol; // final interpolated envelope volume
   int envvoldelta; // delta step between points
   signed char envvolstopped; // flag to say whether envelope has finished or not
   int envpantick; // tick counter for envelope position
   int envpanpos; // envelope position
   int envpanfrac; // fractional interpolated envelope pan
   int envpan; // final interpolated envelope pan
   int envpandelta; // delta step between points
   signed char envpanstopped; // flag to say whether envelope has finished or not
   int fadeoutvol; // volume fade out
   int ivibpos; // instrument vibrato position
   int ivibsweeppos; // instrument vibrato sweep position
   signed char keyoff; // flag whether keyoff has been hit or not)
   unsigned char inst; // last instrument set in channel
   unsigned char realnote; // last realnote set in channel
   unsigned int period; // last period set in channel
   unsigned char recenteffect; // previous row's effect.. used to correct tremolo volume
   unsigned int sampleoffset; // sample offset for this channel in SAMPLES
   unsigned char portadown; // last porta down value (XM)
   unsigned char portaup; // last porta up value (XM)
   unsigned char xtraportadown; // last porta down value (XM)
   unsigned char xtraportaup; // last porta up value (XM)
   unsigned char volslide; // last volume slide value (XM + S3M)
   unsigned char panslide; // pan slide parameter (XM)
   unsigned char retrigx; // last retrig volume slide used (XM + S3M)
   unsigned char retrigy; // last retrig tick count used (XM + S3M)
   int portatarget; // note to porta to
   unsigned char portaspeed; // porta speed
   signed char vibpos; // vibrato position
   unsigned char vibspeed; // vibrato speed
   unsigned char vibdepth; // vibrato depth
   signed char tremolopos; // tremolo position
   unsigned char tremolospeed; // tremolo speed
   unsigned char tremolodepth; // tremolo depth
   unsigned char tremorpos; // tremor position (XM + S3M)
   unsigned char tremoron; // remembered parameters for tremor (XM + S3M)
   unsigned char tremoroff; // remembered parameters for tremor (XM + S3M)
   int patlooprow;
   int patloopno; // pattern loop variables for effect E6x
   unsigned char wavecontrol; // waveform type for vibrato and tremolo (4bits each)
   unsigned char finevslup; // parameter for fine volume slide down
   unsigned char fineportaup; // parameter for fine porta slide up
   unsigned char fineportadown; // parameter for fine porta slide down
}FMUSIC_CHANNEL;

// Song type - contains info on song
typedef struct FMUSIC_MODULE{
   FMUSIC_PATTERN* pattern; // patterns array for this song
   FMUSIC_INSTRUMENT* instrument; // instrument array for this song (not used in MOD/S3M)
   int mixer_samplesleft;
   int mixer_samplespertick;
   int numpatterns; // number of physical patterns
   int numpatternsmem; // number of allocated patterns

   // 6 bytes read here in the loader for size, dont alter order or insert!!!
   short numorders; // number of orders (song length)
   short restart; // restart position
   short numchannels; // number of channels

   // 256+8 bytes read here in the loader for size, dont alter order or insert!!!
   short numinsts; // number of instruments
   unsigned short flags; // flags such as linear frequency, format specific quirks etc
   short defaultspeed;
   short defaultbpm;
   unsigned char orderlist[256]; // pattern playing order list

   int globalvolume; // global mod volume
   int tick; // current mod tick
   int speed; // speed of song in ticks per row
   int row; // current row in pattern
   int order; // current song order position
   int patterndelay; // pattern delay counter
   int nextrow; // current row in pattern
   int nextorder; // current song order position
   int time_ms; // time passed in seconds since song started
   unsigned char globalvsl; // global mod volume
}FMUSIC_MODULE;

// Private function
void __stdcall FMUSIC_SetBPM(FMUSIC_MODULE*,int);

enum FMUSIC_XMCOMMANDS{
   FMUSIC_XM_ARPEGGIO,
   FMUSIC_XM_PORTAUP,
   FMUSIC_XM_PORTADOWN,
   FMUSIC_XM_PORTATO,
   FMUSIC_XM_VIBRATO,
   FMUSIC_XM_PORTATOVOLSLIDE,
   FMUSIC_XM_VIBRATOVOLSLIDE,
   FMUSIC_XM_TREMOLO,
   FMUSIC_XM_SETPANPOSITION,
   FMUSIC_XM_SETSAMPLEOFFSET,
   FMUSIC_XM_VOLUMESLIDE,
   FMUSIC_XM_PATTERNJUMP,
   FMUSIC_XM_SETVOLUME,
   FMUSIC_XM_PATTERNBREAK,
   FMUSIC_XM_SPECIAL,
   FMUSIC_XM_SETSPEED,
   FMUSIC_XM_SETGLOBALVOLUME,
   FMUSIC_XM_GLOBALVOLSLIDE,
   FMUSIC_XM_I,
   FMUSIC_XM_J,
   FMUSIC_XM_KEYOFF,
   FMUSIC_XM_SETENVELOPEPOS,
   FMUSIC_XM_M,
   FMUSIC_XM_N,
   FMUSIC_XM_O,
   FMUSIC_XM_PANSLIDE,
   FMUSIC_XM_Q,
   FMUSIC_XM_MULTIRETRIG,
   FMUSIC_XM_S,
   FMUSIC_XM_TREMOR,
   FMUSIC_XM_U,
   FMUSIC_XM_V,
   FMUSIC_XM_W,
   FMUSIC_XM_EXTRAFINEPORTA,
   FMUSIC_XM_Y,
   FMUSIC_XM_Z,
};

enum FMUSIC_XMCOMMANDSSPECIAL{
   FMUSIC_XM_SETFILTER,
   FMUSIC_XM_FINEPORTAUP,
   FMUSIC_XM_FINEPORTADOWN,
   FMUSIC_XM_SETGLISSANDO,
   FMUSIC_XM_SETVIBRATOWAVE,
   FMUSIC_XM_SETFINETUNE,
   FMUSIC_XM_PATTERNLOOP,
   FMUSIC_XM_SETTREMOLOWAVE,
   FMUSIC_XM_SETPANPOSITION16,
   FMUSIC_XM_RETRIG,
   FMUSIC_XM_FINEVOLUMESLIDEUP,
   FMUSIC_XM_FINEVOLUMESLIDEDOWN,
   FMUSIC_XM_NOTECUT,
   FMUSIC_XM_NOTEDELAY,
   FMUSIC_XM_PATTERNDELAY,
   FMUSIC_XM_FUNKREPEAT,
};

#endif
