// SOUND.H
// -------
// uFMOD public source code release. (MiniFMOD v1.70 based.) Provided as-is.
// Original Visual C++ source optimized and debugged by Asterix and Quantum.
// uFMOD sourcecode is (c)opyleft [WASM.RU]. All rights reserved & reversed.
// MiniFMOD sourcecode is (c)opyright Firelight Technologies (R), 2000-2005.
// uFMOD sourcecode is in no way representative of FMOD 3 source.
// This source must not be redistributed without this notice.

#ifndef _SOUND_H_
#define _SOUND_H_

#include <windows.h>

// DEFINITIONS
#define FSOUND_MIXDIR_FORWARDS 0
#define FSOUND_MixRate 44100 // mixing rate in hz

// Sample description bitfields, OR them together for loading and describing samples
#define FSOUND_LOOP_OFF 0x01 // For non looping samples
#define FSOUND_LOOP_NORMAL 0x02 // For forward looping samples
#define FSOUND_LOOP_BIDI 0x04 // For bidirectional looping samples. (no effect if in hardware)
#define FSOUND_16BITS 0x10 // For 16 bit samples
#define FSOUND_SIGNED 0x100 // For source data containing signed data
#define FSOUND_DELTA 0x200 // For source data stored as delta values

// STRUCTURE DEFINITIONS 

// Sample type - contains info on sample
typedef struct FSOUND_SAMPLE{
   unsigned char* buff; // pointer to sound data

   // DONT CHANGE THE ORDER OF THESE NEXT MEMBERS AS THEY ARE ALL LOADED IN A SINGLE READ
   unsigned int length; // sample length (samples)
   unsigned int loopstart; // sample loop start (samples)
   unsigned int looplen; // sample loop length (samples)
   unsigned char defvol; // sample default volume
   signed char finetune; // finetune value from -128 to 127

   int defpan; // sample default pan
   unsigned char bits; // bits per sample
   unsigned char loopmode; // loop type
   signed char relative; // relative note
}FSOUND_SAMPLE;

// Channel type - contains information on a mixing channel
typedef struct{
   int index; // position in channel pool
   int volume; // current volume (00-FFh)
   int frequency; // speed or rate of playback in hz
   int actualvolume; // driver level current volume
   int actualpan; // driver level panning value
   unsigned int sampleoffset; // sample offset (sample starts playing from here)
   FSOUND_SAMPLE* sptr; // currently playing sample

   // software mixer stuff
   unsigned int leftvolume; // mixing information. adjusted volume for left channel (panning involved)
   unsigned int rightvolume; // mixing information. adjusted volume for right channel (panning involved)
   unsigned int mixpos; // mixing information. high part of 32:32 fractional position in sample
   unsigned int mixposlo; // mixing information. low part of 32:32 fractional position in sample
   unsigned int speedlo; // mixing information. playback rate - low part fractional
   unsigned int speedhi; // mixing information. playback rate - high part fractional
   unsigned int speeddir; // mixing information. playback direction - forwards or backwards

   // software mixer volume ramping stuff
   unsigned int ramp_lefttarget;
   unsigned int ramp_righttarget;
   unsigned int ramp_leftvolume;
   unsigned int ramp_rightvolume;
   unsigned int ramp_leftspeed;
   unsigned int ramp_rightspeed;
   unsigned int ramp_count;
}FSOUND_CHANNEL;

#endif
