; MUSIC.INC
; ---------
; uFMOD public source code release. (MiniFMOD v1.70 based.) Provided as-is.
; Original Visual C++ source optimized and debugged by Asterix and Quantum.
; uFMOD sourcecode is (c)opyleft [WASM.RU]. All rights reserved & reversed.
; MiniFMOD sourcecode is (c)opyright Firelight Technologies (R), 2000-2005.
; uFMOD sourcecode is in no way representative of FMOD 3 source.
; This source must not be redistributed without this notice.

include sound.inc

FMUSIC_TIMMEINFO STRUCT
   order db ?
   row db ?
   ms dd ?
FMUSIC_TIMMEINFO ENDS

; Pattern data type
FMUSIC_PATTERN STRUCT
   rows                 dd ?
   data                 dd ? ; pointer to FMUSIC_NOTE
FMUSIC_PATTERN ENDS

; Multi sample extended instrument
FMUSIC_INSTRUMENT STRUCT
   numsamples           dd ? ; number of samples in this instrument
   sample               dd 16 dup (?) ; 16 pointers to FSOUND_SAMPLE per instrument
   keymap               db 96 dup (?) ; sample keymap assignments
   VOLPoints            dw 24 dup (?) ; volume envelope points
   PANPoints            dw 24 dup (?) ; panning envelope points
   ; 16 bytes read here in the loader for size, dont alter order or insert!!! .:.
   VOLnumpoints         db ? ; number of volume envelope points
   PANnumpoints         db ? ; number of panning envelope points
   VOLsustain           db ? ; volume sustain point
   VOLLoopStart         db ? ; volume envelope loop start
   VOLLoopEnd           db ? ; volume envelope loop end
   PANsustain           db ? ; panning sustain point
   PANLoopStart         db ? ; panning envelope loop start
   PANLoopEnd           db ? ; panning envelope loop end
   VOLtype              db ? ; type of envelope,bit 0:On 1:Sustain 2:Loop
   PANtype              db ? ; type of envelope,bit 0:On 1:Sustain 2:Loop
   VIBtype              db ? ; instrument vibrato type
   VIBsweep             db ? ; time it takes for vibrato to fully kick in
   VIBdepth             db ? ; depth of vibrato
   VIBrate              db ? ; rate of vibrato
   VOLfade              dw ? ; fade out value
   ; .:.
FMUSIC_INSTRUMENT ENDS

; Channel type - contains information on a mod channel
FMUSIC_CHANNEL STRUCT
   note                 db ? ; last note set in channel
   samp                 db ? ; last sample set in channel
   notectrl             db ? ; flags for FSOUND
   cptr                 dd ? ; pointer to FSOUND_CHANNEL system mixing channel
   sptr                 dd ? ; pointer to FSOUND_SAMPLE system sample
   freq                 dd ? ; current mod frequency period for this channel
   volume               dd ? ; current mod volume for this channel
   pan                  dd ? ; current mod pan for this channel
   voldelta             dd ? ; delta for volume commands... tremolo/tremor, etc.
   freqdelta            dd ? ; delta for frequency commands... vibrato/arpeggio, etc.
   envvoltick           dd ? ; tick counter for envelope position
   envvolpos            dd ? ; envelope position
   envvolfrac           dd ? ; fractional interpolated envelope volume
   envvol               dd ? ; final interpolated envelope volume
   envvoldelta          dd ? ; delta step between points
   envvolstopped        db ? ; flag to say whether envelope has finished or not
   envpantick           dd ? ; tick counter for envelope position
   envpanpos            dd ? ; envelope position
   envpanfrac           dd ? ; fractional interpolated envelope pan
   envpan               dd ? ; final interpolated envelope pan
   envpandelta          dd ? ; delta step between points
   envpanstopped        db ? ; flag to say whether envelope has finished or not
   fadeoutvol           dd ? ; volume fade out
   ivibpos              dd ? ; instrument vibrato position
   ivibsweeppos         dd ? ; instrument vibrato sweep position
   keyoff               db ? ; flag whether keyoff has been hit or not)
   inst                 db ? ; last instrument set in channel
   realnote             db ? ; last realnote set in channel
   period               dd ? ; last period set in channel
   recenteffect         db ? ; previous row's effect... used to correct tremolo volume
   sampleoffset         dd ? ; sample offset for this channel in SAMPLES
   portadown            db ? ; last porta down value (XM)
   portaup              db ? ; last porta up value (XM)
   xtraportadown        db ? ; last porta down value (XM)
   xtraportaup          db ? ; last porta up value (XM)
   volslide             db ? ; last volume slide value (XM + S3M)
   panslide             db ? ; pan slide parameter (XM)
   retrigx              db ? ; last retrig volume slide used (XM + S3M)
   retrigy              db ? ; last retrig tick count used (XM + S3M)
   portatarget          dd ? ; note to porta to
   portaspeed           db ? ; porta speed
   vibpos               db ? ; vibrato position
   vibspeed             db ? ; vibrato speed
   vibdepth             db ? ; vibrato depth
   tremolopos           db ? ; tremolo position
   tremolospeed         db ? ; tremolo speed
   tremolodepth         db ? ; tremolo depth
   tremorpos            db ? ; tremor position (XM + S3M)
   tremoron             db ? ; remembered parameters for tremor (XM + S3M)
   tremoroff            db ? ; remembered parameters for tremor (XM + S3M)
   patlooprow           dd ?
   patloopno            dd ? ; pattern loop variables for effect E6x
   wavecontrol          db ? ; waveform type for vibrato and tremolo (4bits each)
   finevslup            db ? ; parameter for fine volume slide down
   fineportaup          db ? ; parameter for fine porta slide up
   fineportadown        db ? ; parameter for fine porta slide down
FMUSIC_CHANNEL ENDS

; Song type - contains info on song
FMUSIC_MODULE STRUCT
   pattern              dd ? ; pointer to FMUSIC_PATTERN array for this song
   instrument           dd ? ; pointer to FMUSIC_INSTRUMENT array for this song (not used in MOD/S3M)
   mixer_samplesleft    dd ?
   mixer_samplespertick dd ?
   numpatterns          dd ? ; number of physical patterns
   numpatternsmem       dd ? ; number of allocated patterns

   ; 6 bytes read here in the loader for size, dont alter order or insert!!! .:.
   numorders            dw ? ; number of orders (song length)
   restart              dw ? ; restart position
   numchannels          dw ? ; number of channels
   ; .:.

   ; 256+8 bytes read here in the loader for size, dont alter order or insert!!! .:.
   numinsts             dw ? ; number of instruments
   flags                dw ? ; flags such as linear frequency, format specific quirks, etc.
   defaultspeed         dw ?
   defaultbpm           dw ?
   orderlist            db 256 dup (?) ; pattern playing order list
   ; .:.

   globalvolume         dd ? ; global mod volume
   tick                 dd ? ; current mod tick
   speed                dd ? ; speed of song in ticks per row
   row                  dd ? ; current row in pattern
   order                dd ? ; current song order position
   patterndelay         dd ? ; pattern delay counter
   nextrow              dd ? ; current row in pattern
   nextorder            dd ? ; current song order position
   time_ms              dd ? ; time passed in seconds since song started
   globalvsl            db ? ; global mod volume
FMUSIC_MODULE ENDS
