; SOUND.INC
; ---------
; uFMOD public source code release. (MiniFMOD v1.70 based.) Provided as-is.
; Original Visual C++ source optimized and debugged by Asterix and Quantum.
; uFMOD sourcecode is (c)opyleft [WASM.RU]. All rights reserved & reversed.
; MiniFMOD sourcecode is (c)opyright Firelight Technologies (R), 2000-2005.
; uFMOD sourcecode is in no way representative of FMOD 3 source.
; This source must not be redistributed without this notice.

include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\winmm.inc

; *** CONSTANTS ***

FSOUND_MIXDIR_FORWARDS  EQU 0
FSOUND_MIXDIR_BACKWARDS EQU 1
FSOUND_MixRate          EQU 44100 ; mixing rate in hz
; Sample description bitfields, OR them together for loading and describing samples
FSOUND_LOOP_OFF         EQU 1     ; non looping samples
FSOUND_LOOP_NORMAL      EQU 2     ; forward looping samples
FSOUND_LOOP_BIDI        EQU 4     ; bidirectional looping samples. (no effect if in hardware)
FSOUND_16BITS           EQU 10h   ; 16 bit samples
FSOUND_SIGNED           EQU 100h  ; source data containing signed data
FSOUND_DELTA            EQU 200h  ; For source data stored as delta values
FSOUND_BlockSize        EQU (FSOUND_MixRate / 50 + 3) AND 0FFFFFFFCh ; Number of *samples*
FSOUND_BufferSizeMs     EQU 1000
totalblocks             EQU FSOUND_BufferSizeMs / 10
FSOUND_BufferSize       EQU FSOUND_BlockSize * totalblocks ; size of 1 'latency' ms buffer in bytes

; *** STRUCTS ***

; Sample type - contains info on sample
FSOUND_SAMPLE STRUCT
   buff                 dd ? ; pointer to sound data

   ; DONT CHANGE THE ORDER OF THESE NEXT MEMBERS
   ; AS THEY ARE ALL LOADED IN A SINGLE READ .:.
   _length              dd ? ; sample length (samples)
   loopstart            dd ? ; sample loop start (samples)
   looplen              dd ? ; sample loop length (samples)
   defvol               db ? ; sample default volume
   finetune             db ? ; finetune value from -128 to 127
   ; .:.

   defpan               dd ? ; sample default pan
   bits                 db ? ; bits per sample
   loopmode             db ? ; loop type
   relative             db ? ; relative note
FSOUND_SAMPLE ENDS

; Channel type - contains information on a mixing channel
FSOUND_CHANNEL STRUCT
   index                dd ? ; position in channel pool
   volume               dd ? ; current volume (00-FFh)
   frequency            dd ? ; speed or rate of playback in hz
   actualvolume         dd ? ; driver level current volume
   actualpan            dd ? ; driver level panning value
   sampleoffset         dd ? ; sample offset (sample starts playing from here)
   sptr                 dd ? ; pointer to FSOUND_SAMPLE currently playing sample
   leftvolume           dd ? ; mixing information. adjusted volume for left channel (panning involved)
   rightvolume          dd ? ; mixing information. adjusted volume for right channel (panning involved)
   mixpos               dd ? ; mixing information. high part of 32:32 fractional position in sample
   mixposlo             dd ? ; mixing information. low part of 32:32 fractional position in sample
   speedlo              dd ? ; mixing information. playback rate - low part fractional
   speedhi              dd ? ; mixing information. playback rate - high part fractional
   speeddir             dd ? ; mixing information. playback direction - forwards or backwards
   ; ramping stuff
   ramp_lefttarget      dd ?
   ramp_righttarget     dd ?
   ramp_leftvolume      dd ?
   ramp_rightvolume     dd ?
   ramp_leftspeed       dd ?
   ramp_rightspeed      dd ?
   ramp_count           dd ?
FSOUND_CHANNEL ENDS

FSOUND_SoundBlock STRUCT
   wavehdr              WAVEHDR <>
   data                 dd ?
FSOUND_SoundBlock ENDS

; *** EXTERNS ***
externdef FSOUND_Channel:FSOUND_CHANNEL ; channel array